/*
Bubbles.c

By:  George A. Nelson on	1  April 1987
							16 January 1987 Converted to LightspeedC
							18 Feb 2005 Converted to Codewarrior Pro4
							26 Feb 2005 Converted to Win32 and MSVC6

Description:  Fast bubbles, inspired by slow Object Logo version I wrote.
	Bubbles push away from each other.  New bubbles are randomly placed,
	and push existing bubbles away.  Bubbles can be dragged.
*/

#include "..\\Common\\CmnHdr.h"
#include <WindowsX.h>
#include <tchar.h>
#include "resource.h"
#include "Bubbles.h"
#include "main.h"
#include "..\\Common\\Utility.h"
#include <stdlib.h>
#include <math.h>


#define TIMERID 1

HCURSOR gArrowCursor, gPinCursor, gWandCursor;

DWORD gTimerIncrement;


/*************************  Bubbles Wind  **************************/


/*	Bubble window stuff:
*/
#define BUBBLESDATAPROP _T("BUBBLESDATA")
#define LMAX 200

enum { Bbl_Make, Bbl_Gravity, Bbl_Design, Bbl_Yeast };
enum { moveMode, popMode, sizeMode };

typedef struct POINTF {
	float		x;
	float		y;
} POINTF;

typedef struct RECTF {
	float		left;
	float		top;
	float		right;
	float		bottom;
} RECTF;

typedef struct Bubble {
	POINTF		loc;
	float		size,
				adultSize;
	short		color,			/* color from list */
				nextBbl;		/* Next bubble in breadth-first avoid list */
	char		yeasting,		/* bubble is a bud growing to full size */
				gravityNow;
} Bubble;

typedef struct BubblesData {
	POINTF		lastMouse;		/* Previous mouse position, for bubble track */
	float		curSize;		/* Make new bubbles this size */
	short		mouseBbl,		/* Currently moused bubble */
				avoidBbl,		/* Current bubble avoiding others, head of list */
				lastBbl,		/* End of breadth-first avoid list */
				nBubbles,		/* Number of bubbles. */
				mode;			/* what does a mouse-click do? */
	BOOL		inADrag,
				design,
				modeByMenu;
	DWORD		nextYeast,		/* time of next yeast-start */
				nextGrow,		/* time to next grow yeast */
				nextGravity;
	Bubble		bubbles[LMAX+1]; /* bubble 0 is special */
} BubblesData;

static COLORREF colors[] = {
	RGB(0xFF,0xFF,0xFF),		// white
	RGB(0,0,0),					// black
	RGB(0xC0,0xC0,0xC0),
	RGB(0x80,0x80,0x80),
	RGB(0x40,0x40,0x40),
	RGB(0xFF,0,0),				// red
	RGB(0,0xFF,0),				// green
	RGB(0,0,0xFF),				// blue
	RGB(0xC0,0xC0,0),
	RGB(0,0xC0,0xC0),
	RGB(0xC0,0,0xC0),
	RGB(0xC0,0xC0,0x40),
	RGB(0x40,0xC0,0xC0),
	RGB(0xC0,0x40,0xC0),
	RGB(0xC0,0xC0,0x80),
	RGB(0x80,0xC0,0xC0),
	RGB(0xC0,0x80,0xC0),
};


/*	Add bubble to Avoid List if not already in list.
*/
	static void
Enlist( BubblesData * bdp, short b1 )
{
	if( ! b1 )
		return;

	Bubble *bp = &bdp->bubbles[ b1 ];
	
	if( bp->nextBbl == -1 ) {				/* is bubble free? */
		bp->nextBbl = 0;
		if( bdp->lastBbl > 0 )
			bdp->bubbles[ bdp->lastBbl ].nextBbl = b1;
		bdp->lastBbl = b1;
		
		if( bdp->avoidBbl <= 0 )
			bdp->avoidBbl = b1;
	}
}


	static short
Delist( BubblesData * bdp )
{
	short b1 = bdp->avoidBbl;
	
	if( b1 > 0 ) {
		Bubble *bp = &bdp->bubbles[ b1 ];
		
		bdp->avoidBbl = bp->nextBbl;
		bp->nextBbl = -1;					/* bubble is free again */
		
		if( bdp->lastBbl == b1 )
			bdp->lastBbl = 0;
	}
	
	return b1;
}


	static void
EnlistAllBubbles( BubblesData * bdp )
{
	if( ! bdp->nBubbles ) {
		bdp->lastBbl = 0;
		bdp->avoidBbl = 0;
		return;
	}

	for( short b1 = 1;  b1 <= bdp->nBubbles;  ++b1 )
		bdp->bubbles[b1].nextBbl = (short) (b1-1);
	bdp->lastBbl = 1;
	bdp->avoidBbl = bdp->nBubbles;
}


/*	Draw a bubble.
*/
	static void
DrawBubble( HDC hdc, BubblesData * bdp, short b1 )
{
	Bubble *bp = &bdp->bubbles[ b1 ];
	
	RECTF r;
	r.left   = bp->loc.x - bp->size;
	r.top    = bp->loc.y - bp->size;
	r.right  = bp->loc.x + bp->size;
	r.bottom = bp->loc.y + bp->size;
	
	int oldROP = SetROP2( hdc, R2_XORPEN );
	HPEN oldPen = SelectPen( hdc, CreatePen(PS_SOLID, 2, RGB(0x80,0x80,0x80)) );
	HBRUSH oldBrush = SelectBrush( hdc, (bdp->design) ?
		CreateSolidBrush(colors[bp->color]) : GetStockObject(NULL_BRUSH) );
	
	Ellipse( hdc, (int) r.left, (int) r.top, (int) r.right, (int) r.bottom );
	
	DeleteObject( SelectObject(hdc, oldBrush) );
	DeleteObject( SelectObject(hdc, oldPen) );
	SetROP2( hdc, oldROP );
}


	static void
DrawAllBubbles( HDC hdc, BubblesData * bdp )
{
	for( short b1 = 1;  b1 <= bdp->nBubbles;  ++b1 )
		DrawBubble( hdc, bdp, b1 );
}


	static void
MoveBubble( HDC hdc, BubblesData * bdp, short b1, POINTF const & move )
{
	if( move.x*move.x + move.y*move.y > 1/256.0	/* did we move (more than 1/16th)? */
	 && b1 != bdp->mouseBbl )					/* don't mess with bubble being moused on */
	{
		Enlist( bdp, b1 );
		
		DrawBubble( hdc, bdp, b1 );
		
		Bubble *bp = &bdp->bubbles[ b1 ];
		bp->loc.x += move.x;
		bp->loc.y += move.y;
		
		DrawBubble( hdc, bdp, b1 );
	}
}


/*	Make a new bubble, placing it at the end of the bubble list (expand
	the bubble list by 1).  The position is randomly determined, within
	the window boundaries.
*/
	static void
MakeBubble( HWND hwnd, HDC hdc, BubblesData * bdp )
{
	if( bdp->nBubbles < LMAX-1 ) {
		short b1 = ++bdp->nBubbles;
		Bubble *bp = &bdp->bubbles[ b1 ];
		
		RECT r;
		GetClientRect( hwnd, &r );
		bp->loc.x      = (float) RandInt( r.right );
		bp->loc.y      = (float) RandInt( r.bottom );
		bp->size = bp->adultSize = bdp->curSize;
		bp->color      = (short) RandInt( chDIMOF(colors) );
		bp->nextBbl    = -1;
		bp->yeasting   = FALSE;
		bp->gravityNow = FALSE;
		
		Enlist( bdp, b1 );
		DrawBubble( hdc, bdp, b1 );
	}
}


	static void
EraseAllBubbles( BubblesData * bdp )
{
	bdp->nBubbles = 0;
	EnlistAllBubbles( bdp );
}


	static void
Pop1Bubble( HDC hdc, BubblesData * bdp )
{
	if( bdp->mouseBbl > 0 ) {
		DrawBubble( hdc, bdp, bdp->mouseBbl );
		Bubble *bp = &bdp->bubbles[ bdp->mouseBbl ];
		memmove( bp, bp+1, (bdp->nBubbles-- - bdp->mouseBbl)*sizeof(Bubble) );
		EnlistAllBubbles( bdp );
		bdp->mouseBbl = 0;
	}
}


	static void
Overlap( BubblesData * bdp, POINTF * lapp, short b1, short b2 )
{
	Bubble *b1p = &bdp->bubbles[ b1 ];
	Bubble *b2p = &bdp->bubbles[ b2 ];
	
	POINTF dif = { b2p->loc.x - b1p->loc.x, b2p->loc.y - b1p->loc.y };
	double dist2 = dif.x*dif.x + dif.y*dif.y;
	double fullSize = b1p->size + b2p->size;
	double fullSize2 = fullSize*fullSize;
	
	POINTF lap = { 0.0, 0.0 };
	
	if( dist2 < fullSize2 ) {
		double dist = sqrt( dist2 );
		
		if( dist < 0.125 ) {
			lap.x = (float) fullSize;
			lap.y = 0.0;
		}
		else {
			double overlap = fullSize - dist;
			lap.x = (float) (dif.x * overlap / dist); //Scale( dif.h, overlap, dist );
			lap.y = (float) (dif.y * overlap / dist); //Scale( dif.v, overlap, dist );
		}
	}
	
	*lapp = lap;
}


/*	Drag a bubble
*/
	static void
FindClick( BubblesData * bdp, POINTF const & at )
{
	bdp->bubbles[0].loc.x = at.x;
	bdp->bubbles[0].loc.y = at.y;
	bdp->mouseBbl = 0;
	
	for( short b1 = 1;  b1 <= bdp->nBubbles;  ++b1 ) {
		POINTF lap;
		Overlap( bdp, &lap, 0, b1 );
		if( lap.x || lap.y ) {
			bdp->mouseBbl = b1;
			bdp->lastMouse = at;
			bdp->inADrag = TRUE;
			break;
		}
	}
}


	static void
DragBubble( HDC hdc, BubblesData * bdp, POINTF const & at )
{
	if( bdp->mouseBbl > 0 ) {
		POINTF move = { at.x - bdp->lastMouse.x, at.y - bdp->lastMouse.y };
		bdp->lastMouse = at;
		
		short b1 = bdp->mouseBbl;			/* mouseBbl immune to Move() during Avoid()... */
		bdp->mouseBbl = 0;					/* ...so it doesn't jerk about when it hits... */
		MoveBubble( hdc, bdp, b1, move );	/* ...other bubbles during a drag */
		bdp->mouseBbl = b1;
		Enlist( bdp, b1 );
	}
}


	static void
GrowBubble( HDC hdc, BubblesData * bdp, short b1, float size )
{
	Bubble *bp = &bdp->bubbles[ b1 ];
	
	if( size != bp->size ) {
		DrawBubble( hdc, bdp, b1 );
		bp->size = size;
		DrawBubble( hdc, bdp, b1 );
	}
	
	Enlist( bdp, b1 );
}


	static void
SizeBubble( HDC hdc, BubblesData * bdp, POINTF const & at )
{
	if( bdp->mouseBbl > 0 ) {
		short b1 = bdp->mouseBbl;
		Bubble *bp = &bdp->bubbles[ b1 ];
		
		POINTF dif;
		dif.x = at.x - bp->loc.x;
		dif.y = at.y - bp->loc.y;
		bdp->lastMouse = at;					/* just to be polite */
		float size = (float) sqrt( dif.x*dif.x + dif.y*dif.y );
		
		GrowBubble( hdc, bdp, b1, size );
		bdp->bubbles[b1].adultSize = size;
		bdp->curSize = size;
	}
}


	static void
NewYeast( HDC hdc, BubblesData * bdp, short b1 )
{
	if( bdp->nBubbles  &&  bdp->nBubbles < LMAX-1 ) {
		Bubble *bp = &bdp->bubbles[ b1 ];
		
		if( bp->size < bp->adultSize/2.0 )
			return;							/* too young */
		
		POINTF offset;
		offset.x = (float) RandInt( (int) bp->size );	/* pick a spot on the circumference */
		offset.y = (float) sqrt( bp->size*bp->size - offset.x*offset.x );
		
		/* pick a quadrant */
		POINTF loc;
		loc.x = bp->loc.x + (RandInt(2) ? offset.x : -offset.x);
		loc.y = bp->loc.y + (RandInt(2) ? offset.y : -offset.y);
		
		short color = bp->color;
		float adultSize = bp->adultSize;
		
		/*	Create the new bubble from the parent bubble.
		*/		
		short b2 = ++bdp->nBubbles;
		bp = &bdp->bubbles[ b2 ];
		
		bp->loc        = loc;
		bp->size       = 1;					/* initially very small */
		bp->adultSize  = adultSize;
		bp->nextBbl    = -1;				/* not in list */
		bp->color      = color;				/* inherit color of parent */
		bp->yeasting   = TRUE;
		bp->gravityNow = FALSE;
		
		Enlist( bdp, b2 );
		DrawBubble( hdc, bdp, b2 );
	}
}


	static void
DoYeast( HDC hdc, BubblesData * bdp )
{
	DWORD tickCount = GetTickCount();

	/*	Start some yeast.
	*/
	if( bdp->nextYeast ) {
		if( ((long) (bdp->nextYeast - tickCount)) < 0 ) {
			bdp->nextYeast = tickCount + 16*RandInt( 10 );
			
			for( short i = 1;  i <= bdp->nBubbles;  ++ i ) {
				if( RandInt(1000) < 3 )
					NewYeast( hdc, bdp, i );
			}
			
			if( !bdp->nextGrow )
				bdp->nextGrow = tickCount;
		}
	}
	
	/*	Grow new bubbles.
	*/
	if( bdp->nextGrow  &&  ((long) (bdp->nextGrow - tickCount)) < 0 ) {
		bdp->nextGrow = tickCount + 16*20;
		
		BOOL grewOne = FALSE;
		
		for( short i = 1;  i <= bdp->nBubbles;  ++ i ) {
			Bubble *bp = &bdp->bubbles[ i ];
			if( bp->yeasting ) {
				grewOne = TRUE;
				if( bp->size < bp->adultSize )
					GrowBubble( hdc, bdp, i, bp->size + 1.0f );
				else
					bp->yeasting = FALSE;
			}
		}
		
		if( !grewOne )
			bdp->nextGrow = 0;
	}
}


	static void
Gravity( BubblesData * bdp )
{
short				b1;

	for( b1 = 1;  b1 <= bdp->nBubbles;  ++b1 ) {
		Bubble *bp = &bdp->bubbles[b1];
		if( !bp->gravityNow ) {
			bp->gravityNow = TRUE;
			Enlist( bdp, b1 );
		}
	}
}


/*	Make the next bubble avoid its neighbors.  When the end of the bubble
	list is reached, return to the front of the list.  An avoid requires a
	scan of the entire list checking each other bubble's distance (squared);
	if an other bubble is too close, add the overlap into the amount to
	be pushed (vector add).  Bubbles also avoid the walls.
*/
	static void
Avoid1Bubble( HWND hwnd, HDC hdc, BubblesData * bdp, short b1 )
{
	if( b1 <= 0 )							/* nothing in list? */
		return;
	
	Bubble *bp = &bdp->bubbles[ b1 ];
	
	/*	Move other bubbles out of our way.
	*/
	for( short b2 = 1;  b2 <= bdp->nBubbles;  ++b2 ) {
		if( b2 != b1 ) {
			POINTF lap;
			Overlap( bdp, &lap, b1, b2 );
			MoveBubble( hdc, bdp, b2, lap );
		}
	}
	
	/*	Make this bubble avoid walls.
	*/
	RECT r;
	GetClientRect( hwnd, &r );
	POINTF meanLap = { 0.0, 0.0 };
	if( bp->gravityNow ) {
		bp->gravityNow = FALSE;
		meanLap.y = 5.0;
	}
	
	POINTF lap = bp->loc;

	if( lap.x < 0.0 )								/* left edge */
		lap.x = 0.0;
	lap.x -= bp->size;
	if( lap.x < 0.0 )
		meanLap.x -= lap.x;
	
	if( lap.y < 0.0 )								/* top edge */
		lap.y = 0.0;
	lap.y -= bp->size;
	if( lap.y < 0.0 )
		meanLap.y -= lap.y;
	
	lap.x = bp->loc.x - r.right;				/* right edge */
	if( lap.x > 0.0 )
		lap.x = 0.0;
	lap.x += bp->size;
	if( lap.x > 0.0 )
		meanLap.x -= lap.x;
	
	lap.y = bp->loc.y - r.bottom;				/* bottom edge */
	if( lap.y > 0.0 )
		lap.y = 0.0;
	lap.y += bp->size;
	if( lap.y > 0.0 )
		meanLap.y -= lap.y;
	
	MoveBubble( hdc, bdp, b1, meanLap );
}


	static BOOL
Cls_OnCreate( HWND hwnd, LPCREATESTRUCT csp )
{
	if( ! gArrowCursor ) {
		gArrowCursor  = LoadCursor( NULL, IDC_ARROW );
		gPinCursor    = LoadCursor( hInst, MAKEINTRESOURCE(IDC_CURSORPIN) );
		gWandCursor   = LoadCursor( hInst, MAKEINTRESOURCE(IDC_CURSORWAND) );

		DWORD ta;
		BOOL d;
		GetSystemTimeAdjustment( &ta, &gTimerIncrement, &d );
		gTimerIncrement /= 10000;
	}

	MDICREATESTRUCT *mcsp = (MDICREATESTRUCT *) csp->lpCreateParams;
	HWND cw = (HWND) mcsp->lParam;

	BubblesData *bdp = (BubblesData *) calloc( sizeof(BubblesData), 1 );
	SetProp( hwnd, BUBBLESDATAPROP, (HANDLE) bdp );
	bdp->curSize     = 20.0;
	bdp->nextGravity = 0;
	bdp->design      = TRUE;
	bdp->nextYeast   = GetTickCount();

	BubblesData *cwbdp = NULL;

	if( IsWindow(cw) ) {
		cwbdp = (BubblesData *) GetProp( cw, BUBBLESDATAPROP );
		
		if( cwbdp ) {
			bdp->curSize     = cwbdp->curSize;
			bdp->nextGravity = cwbdp->nextGravity;
			bdp->design      = cwbdp->design;
			bdp->nextYeast   = cwbdp->nextYeast;
		}
	}

	SetTimer( hwnd, TIMERID, 16, NULL );

	return true;
}


	static void
Cls_OnInitMenuPopup( HWND hwnd, HMENU hmenu, UINT item, BOOL fSystemMenu )
{
	if( hmenu == gFileMenu ) {
		EnableMenuItem( hmenu, ID_FILE_CLOSE, MF_BYCOMMAND | MF_ENABLED );
	}

	if( hmenu == gBubblesMenu ) {
		BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );
		
		if( bdp->nBubbles < LMAX-1 )
			EnableMenuItem( hmenu, ID_BUBBLES_MAKEBUBBLE, MF_BYCOMMAND | MF_ENABLED );
		if( bdp->nBubbles )
			EnableMenuItem( hmenu, ID_BUBBLES_ERASEALLBUBBLES, MF_BYCOMMAND | MF_ENABLED );

		EnableMenuItem( hmenu, ID_BUBBLES_GRAVITY, MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_BUBBLES_DESIGN,  MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_BUBBLES_YEAST,   MF_BYCOMMAND | MF_ENABLED );

		CheckMenuItem( hmenu, ID_BUBBLES_GRAVITY, MF_BYCOMMAND
			| (bdp->nextGravity ? MF_CHECKED       : MF_UNCHECKED) );
		CheckMenuItem( hmenu, ID_BUBBLES_DESIGN,  MF_BYCOMMAND
			| (bdp->design      ? MF_CHECKED       : MF_UNCHECKED) );
		CheckMenuItem( hmenu, ID_BUBBLES_YEAST,   MF_BYCOMMAND
			| (bdp->nextYeast   ? MF_CHECKED       : MF_UNCHECKED) );

		EnableMenuItem( hmenu, ID_BUBBLES_DRAG,   MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_BUBBLES_POP,    MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_BUBBLES_GROW,   MF_BYCOMMAND | MF_ENABLED );

		CheckMenuRadioItem( hmenu, ID_BUBBLES_DRAG, ID_BUBBLES_GROW,
			bdp->mode - moveMode + ID_BUBBLES_DRAG, MF_BYCOMMAND );
	}
}


	static void
Cls_OnCommand( HWND hwnd, int id, HWND hwndCtl, UINT codeNotify )
{
	BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );

	switch( id )
	{
	case ID_BUBBLES_MAKEBUBBLE:
	{
		HDC hdc = GetDC( hwnd );
		MakeBubble( hwnd, hdc, bdp );
		ReleaseDC( hwnd, hdc );
		break;
	}

	case ID_BUBBLES_ERASEALLBUBBLES:
		EraseAllBubbles( bdp );
		InvalidateRect( hwnd, NULL, TRUE );
		break;

	case ID_BUBBLES_GRAVITY:
		if( ! bdp->nextGravity )
			bdp->nextGravity = GetTickCount();		/* turn gravity on */
		else
			bdp->nextGravity = 0;
		break;
	
	case ID_BUBBLES_DESIGN:
		bdp->design = ! bdp->design;
		InvalidateRect( hwnd, NULL, TRUE );
		break;
	
	case ID_BUBBLES_YEAST:
		if( ! bdp->nextYeast )
			bdp->nextYeast = GetTickCount();		/* turn yeasting on */
		else
			bdp->nextYeast = 0;
		break;

	case ID_BUBBLES_DRAG:
	case ID_BUBBLES_POP:
	case ID_BUBBLES_GROW:
		bdp->mode = (short) (id - ID_BUBBLES_DRAG + moveMode);
		bdp->modeByMenu = TRUE;
		break;
	}
}


	static void
SetTheCursor( HWND hwnd, BOOL setIt )
{
	BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );
	
	if( ! bdp->inADrag  &&  ! bdp->modeByMenu ) {
		if( GetKeyState(VK_SHIFT) < 0 )
			bdp->mode = popMode;
		else if( GetKeyState(VK_CONTROL) < 0 )
			bdp->mode = sizeMode;
		else
			bdp->mode = moveMode;
	}

	if( setIt ) {
		switch( bdp->mode ) {
		case moveMode:
			SetCursor( gArrowCursor );
			break;

		case popMode:
			SetCursor( gPinCursor );
			break;

		case sizeMode:
			SetCursor( gWandCursor );
			break;
		}
	}
}


	static BOOL
Cls_OnSetCursor( HWND hwnd, HWND hwndCursor, UINT codeHitTest, UINT msg )
{
	if( hwndCursor == hwnd  &&  codeHitTest == HTCLIENT ) {
		SetTheCursor( hwnd, TRUE );
		return TRUE;
	}

	return FORWARD_WM_SETCURSOR( hwnd, hwndCursor, codeHitTest, msg, DefMDIChildProc );
}


	static void
Cls_OnLButtonDown( HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags )
{
	BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );

	POINTF at = { (float) x, (float) y };
	FindClick( bdp, at );

	if( bdp->mode == popMode ) {
		HDC hdc = GetDC( hwnd );
		Pop1Bubble( hdc, bdp );
		ReleaseDC( hwnd, hdc );
	}
	else {
		SetCapture( hwnd );
	}
}


	static void
Cls_OnMouseMove( HWND hwnd, int x, int y, UINT keyFlags )
{
	BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );
	
	if( bdp->mouseBbl ) {
		switch( bdp->mode ) {
		case moveMode:
		{
			POINTF at = { (float) x, (float) y };
			HDC hdc = GetDC( hwnd );
			DragBubble( hdc, bdp, at );
			ReleaseDC( hwnd, hdc );
			break;
		}

		case popMode:
			break;

		case sizeMode:
		{
			POINTF at = { (float) x, (float) y };
			HDC hdc = GetDC( hwnd );
			SizeBubble( hdc, bdp, at );
			ReleaseDC( hwnd, hdc );
			break;
		}
		}
	}
}


	static void
EndTracking( HWND hwnd )
{
	BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );

	Enlist( bdp, bdp->mouseBbl );
	bdp->mouseBbl = 0;
	bdp->inADrag = FALSE;
}


	static void
Cls_OnLButtonUp( HWND hwnd, int x, int y, UINT keyFlags )
{
	if( GetCapture() == hwnd )
		ReleaseCapture();

	EndTracking( hwnd );
}


	static void
Cls_OnKey( HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags )
{
	if( cRepeat == 1  &&  (vk == VK_SHIFT  ||  vk == VK_CONTROL) ) {
		BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );
		
		bdp->modeByMenu = FALSE;
		
		BOOL setIt = FALSE;
		DWORD atd = GetMessagePos();
		POINT at = { GET_X_LPARAM(atd), GET_Y_LPARAM(atd) };
		if( WindowFromPoint(at) == hwnd ) {
			LRESULT codeHitTest = SendMessage( hwnd, WM_NCHITTEST, 0, atd );
			if( codeHitTest == HTCLIENT )
				setIt = TRUE;
		}
		SetTheCursor( hwnd, setIt );
	}
}


	static void
Cls_OnPaint( HWND hwnd )
{
	BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );

	PAINTSTRUCT ps;
	HDC hdc = BeginPaint( hwnd, &ps );
	DrawAllBubbles( hdc, bdp );
	EndPaint( hwnd, &ps );
}


	static void
Cls_OnTimer( HWND hwnd, UINT id )
{
	if( id == TIMERID ) {
		DWORD tickCount = GetTickCount();
		BubblesData *bdp = (BubblesData *) GetProp( hwnd, BUBBLESDATAPROP );

		if( bdp->nextGravity
		 && ((long) (bdp->nextGravity - tickCount)) < 0 )
		{
			bdp->nextGravity = tickCount + 16*2;
			Gravity( bdp );
		}

		HDC hdc = GetDC( hwnd );
		
		for( DWORD i = tickCount + 16;
		     ((long) (i - GetTickCount())) > 0  &&  bdp->avoidBbl; )
		{
			Avoid1Bubble( hwnd, hdc, bdp, Delist(bdp) );
		}

		DoYeast( hdc, bdp );
		
		ReleaseDC( hwnd, hdc );
	}
}


	static void
Cls_OnDestroy( HWND hwnd )
{
	KillTimer( hwnd, TIMERID );
	SendMessage( mainW, wmuChildClosing, (WPARAM) hwnd, 0 );
	BubblesData *bdp = (BubblesData *) RemoveProp( hwnd, BUBBLESDATAPROP );
	if( bdp )
		free( (void *) bdp );
}


	static BOOL
DoCmdLineArg( HWND hwnd, TCHAR * arg )
{
	int len = strlen( arg );

	if( _tcsncicmp(arg, _T("/make"), len) == 0 )
			SendMessage( hwnd, WM_COMMAND, ID_BUBBLES_MAKEBUBBLE, 0 );
	else if( _tcsncicmp(arg, _T("/gravity"), len) == 0 )
			SendMessage( hwnd, WM_COMMAND, ID_BUBBLES_GRAVITY, 0 );
	else if( _tcsncicmp(arg, _T("/design"), len) == 0 )
			SendMessage( hwnd, WM_COMMAND, ID_BUBBLES_DESIGN, 0 );
	else if( _tcsncicmp(arg, _T("/yeast"), len) == 0 )
			SendMessage( hwnd, WM_COMMAND, ID_BUBBLES_YEAST, 0 );
	else
		return FALSE;

	return TRUE;
}


	LRESULT CALLBACK
BubblesWndProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	HANDLE_MSG( hwnd, WM_CREATE,        Cls_OnCreate );
	HANDLE_MSG( hwnd, WM_COMMAND,       Cls_OnCommand );
	HANDLE_MSG( hwnd, WM_INITMENUPOPUP, Cls_OnInitMenuPopup );
	HANDLE_MSG( hwnd, WM_SETCURSOR,	    Cls_OnSetCursor );
	HANDLE_MSG( hwnd, WM_LBUTTONDOWN,   Cls_OnLButtonDown );
	HANDLE_MSG( hwnd, WM_MOUSEMOVE,	    Cls_OnMouseMove );
	HANDLE_MSG( hwnd, WM_LBUTTONUP,     Cls_OnLButtonUp );
	HANDLE_MSG( hwnd, WM_KEYDOWN,       Cls_OnKey );
	HANDLE_MSG( hwnd, WM_KEYUP,         Cls_OnKey );
	HANDLE_MSG( hwnd, WM_PAINT,	        Cls_OnPaint );
	HANDLE_MSG( hwnd, WM_TIMER,         Cls_OnTimer );
	HANDLE_MSG( hwnd, WM_DESTROY,       Cls_OnDestroy );

	case WM_MDIACTIVATE:
		PostMessage( mainW, wmuChildActivate, 0, 0 );
		EndTracking( hwnd );
		break;

	case WM_SIZE:
		EnlistAllBubbles( (BubblesData *) GetProp(hwnd, BUBBLESDATAPROP) );
		break;

	case WM_CAPTURECHANGED:
		EndTracking( hwnd );
		break;

	case wmuCmdLineArg:
		return DoCmdLineArg( hwnd, (TCHAR *) wParam );
	}

	return DefMDIChildProc( hwnd, uMsg, wParam, lParam );
}
