/*
Lines.c							Copyright 1985, 1987 George A. Nelson
								Permission for use granted provided
								this notice remains intact.

By:  George A. Nelson on	22 March 1985
							16 January 1987 Converted to LightspeedC
							18 April 1988  removed _DrawMenuBar calls
							19 Feb 2005 Converted to Codewarrior Pro4
							24 Feb 2005 Converted to Win32 and MSVC6

Description:  Creates and moves windows, and displays the moving lines in
	a window.  When a pattern is ended, makes a new pattern.  Update
	events are not handled; the window remains trashed until the pattern
	ends and a new pattern starts -- the window is erased at that time.
	
	Would be slicker if the pattern were sized to the window when it is
	drawn (but erasing would suffer).
*/

#include "..\\Common\\CmnHdr.h"
#include <WindowsX.h>
#include <tchar.h>
#include "resource.h"
#include "Lines.h"
#include "main.h"
#include "..\\Common\\Utility.h"
#include <stdlib.h>
#include <math.h>


#define TIMERID 1

#define SMAX 256								/* Must be a power of 2 and a multiple of 4. */
float				sine[SMAX+1];				/* size plus a guard */
const double PI = 3.14159265358979323846;


/*	Convert from HSV to RGB, from <http://www.easyrgb.com/math.html>, which
	recommends against using HSV as a general color space.  HSV is represented
	as a COLORREF, with values 0.255 rather than 0.0..1.0.
*/
	static COLORREF /* RGB */
HSV2RGB( COLORREF hsv )
{
	int R, G, B;
	double H = GetRValue( hsv ) / 255.0;
	double S = GetGValue( hsv ) / 255.0;
	double V = GetBValue( hsv ) / 255.0;

	if ( S == 0.0 )							//HSV values = From 0 to 1
	{
		R = (int) (V * 255.0);				//RGB results = From 0 to 255
		G = (int) (V * 255.0);
		B = (int) (V * 255.0);
	}
	else {
		double var_h = H * 6.0;
		double var_i = floor( var_h );		//Or ... var_i = int( var_h )
		double var_1 = V * ( 1.0 - S );
		double var_2 = V * ( 1.0 - S * ( var_h - var_i ) );
		double var_3 = V * ( 1.0 - S * ( 1 - ( var_h - var_i ) ) );

		double var_r, var_g, var_b;
		if      ( var_i == 0 ) { var_r = V     ; var_g = var_3 ; var_b = var_1 ; }
		else if ( var_i == 1 ) { var_r = var_2 ; var_g = V     ; var_b = var_1 ; }
		else if ( var_i == 2 ) { var_r = var_1 ; var_g = V     ; var_b = var_3 ; }
		else if ( var_i == 3 ) { var_r = var_1 ; var_g = var_2 ; var_b = V     ; }
		else if ( var_i == 4 ) { var_r = var_3 ; var_g = var_1 ; var_b = V     ; }
		else                   { var_r = V     ; var_g = var_1 ; var_b = var_2 ; }

		R = (int) (var_r * 255.0);			//RGB results = From 0 to 255
		G = (int) (var_g * 255.0);
		B = (int) (var_b * 255.0);
	}

	return RGB(R,G,B);
}


/**************************  Lines Wind  ***************************/


/* Lines window stuff:
*/
#define LINESDATAPROP _T("LINESDATA")
#define LMAX 200

enum { Ctrl_Line, Ctrl_Rect, Ctrl_Oval, Ctrl_Lissajous, Ctrl_Plot };

typedef struct LinesData {
	RECT		lines[LMAX];	// Endpoints of lines for this pattern, for erasing.
	COLORREF	colors[LMAX];	// Color of each line (RGB).
	short		patLines,		// Number of lines in this pattern.
				stopLine,		// Number of lines to draw.
				drawLine,		// The current line to draw.
				eraseLine,		// The current line to erase.
				drawMode,		// Type of object to draw.
				delay,
				delayCnt;
	BOOL		eraseBehind,	// Enables erasing.
				repeating;		// Cycle the existing pattern or get a new one.
} LinesData;


/*	Make a new pattern for a window.  All coordinates and scaling factors
	are positive integers, but the sine table contains fractions.  Each
	pattern is based on two Lissajous figures, formed by four harmonically
	related sine waves.

	A sine wave is completely specified by four independent variables; we use
	magnitude, center (or average) value, initial phase, and rate (as a
	multiple of the fundamental period).  Each quantity is randomly determined,
	but is constrained to reasonable values.  The waves should show in the
	window, so the magnitude and center are computed from extrema chosen to be
	within the window.  Since small magnitudes tend to make solid black on the
	screen, they are goosed up.  Otherwise, they, and the initial phase, are
	evenly distributed over the range.

	The rate is weighted so that equal and low harmonic values are more
	probable.  Higher harmonics are given more lines of display so their
	patterns show better.

	To improve the appearance of motion, normally the earlier lines of the
	figure are erased as new ones are drawn; the number of lines showing is
	also randomly determined and evenly distributed.
*/
#pragma warning( disable : 4244 )
 
	static void
MakePattern( HWND hwnd )
{
	LinesData		*ldp = (LinesData *) GetProp( hwnd, LINESDATAPROP );
	long			i;
	long			height, width;				/* window size */
	short			x[8], y[8], phase[8], rate[8], baseRate,
					patLines, patMax, minAmp, mag;
	static char		skew[12] = { 1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 4, 5 };	/* adjusts probabilities */
	double			a0, a1, a2, a3, a4, a5, a6,
					d0, d1, d2, d3, d4, d5, d6;
	
	
	/*	Determine parameters.  Give the pattern a pleasing length, set of
		harmonic relationships, and number of repetitions.  The skew array
		helps by "normalizing" several crucial random quantities.
	*/
	RECT r;
	GetClientRect( hwnd, &r );
	height = r.bottom;
	width = r.right;
	
	baseRate = skew[ RandInt(9) ];
	patMax = (baseRate > 1) ? LMAX : LMAX/2;
	minAmp = ((height < width) ? height : width) / 4;
	if( minAmp > 20 ) minAmp = 20;
	
	patLines = ldp->patLines = patMax/8 + RandInt( patMax - patMax/8 );
	ldp->stopLine  = patLines * skew[ RandInt(9) ];
	ldp->drawLine  = 0;
	ldp->eraseLine = -1  -  RandInt(patLines - 1) / ((baseRate == 1) ? 1 : 2);
	ldp->delay     = skew[ RandInt(9) ];
	
	if( ! ldp->eraseBehind )					/* if we aren't erasing... */
		ldp->stopLine += patLines;				/* ...pause after drawing */
	
	
	/*	Make new wave limits.
	*/
	for( i=0;  i<4;  ++i ) {
		y[i] = RandInt( height );
		x[i] = RandInt( width );
		phase[i] = RandInt( SMAX );
		rate[i]  = 256 * (baseRate + skew[RandInt(12)]);
	}
	for( i=4;  i<8;  ++i ) {
		y[i] = 0x40 + RandInt( 0xC0 );
		x[i] = RandInt( 0x100 );
		phase[i] = RandInt( SMAX );
		rate[i]  = 256 * (baseRate + skew[RandInt(12)]);
	}
	
	
	/*	Convert limits to center, magnitude.  Avoid very small magnitudes.
	*/
	for( i=0;  i<8;  i+=2 ) {					/* do topLeft, botRight */
		mag = x[i]-x[i+1];						/* 2 * magnitude */
		x[i+1] = mag/2 + x[i+1];				/* center */
		if( mag < 0 ) mag = -mag;				/* Make magnitude positive. */
		if( mag < minAmp ) mag = minAmp;		/* No little tiny squiggles. */
		x[i] = mag;
		
		mag = y[i]-y[i+1];
		y[i+1] = mag/2 + y[i+1];
		if( mag < 0 ) mag = -mag;
		if( mag < minAmp ) mag = minAmp;
		y[i] = mag;
	}
	
	
	/*	Fill in the table of line endpoints by stepping each wave through the
		sine table.  a0, etc., accumulate the instantaneous phase.  Since the
		sine table has 256 entries, it is indexed mod 256.  Note that sine table
		entries must be halved to compensate for the magnitude of the wave being
		twice normal (for historical reasons).
	*/
	a0 = phase[0];
	a1 = phase[1];
	a2 = phase[2];
	a3 = phase[3];
	a4 = phase[4];
	a5 = phase[5];
	a6 = phase[6];
	
	double dpatLines = patLines;
	d0 = rate[0] / dpatLines;			/* rate * table size / # lines */
	d1 = rate[1] / dpatLines;
	d2 = rate[2] / dpatLines;
	d3 = rate[3] / dpatLines;
	d4 = rate[4] / dpatLines;
	d5 = rate[5] / dpatLines;
	d6 = rate[6] / dpatLines;
	
	
	for( i = 0;  i < patLines;  ++i ) {
		ldp->lines[i].left   = x[1] + x[0]*sine[((int) a0) % 256]/2;
		ldp->lines[i].top    = y[1] + y[0]*sine[((int) a1) % 256]/2;
		ldp->lines[i].right  = x[3] + x[2]*sine[((int) a2) % 256]/2;
		ldp->lines[i].bottom = y[3] + y[2]*sine[((int) a3) % 256]/2;
		
		int H = x[5] + x[4]*sine[((int) a4) % 256]/2;
		int S = y[5] + y[0]*sine[((int) a5) % 256]/2;
		int V = y[7] + y[6]*sine[((int) a6) % 256]/2;
		ldp->colors[i] = HSV2RGB( RGB(H,S,V) );

		a0 += d0;
		a1 += d1;
		a2 += d2;
		a3 += d3;
		a4 += d4;
		a5 += d5;
		a6 += d6;
	}
}


/*	Draw the current instance of the current type of object in the window.
	Use the current PenState.
*/
	static void
DrawIt( HDC hdc, LinesData * ldp, short i )
{
	RECT			*rp, *oldRp, sR;
	
	rp = &ldp->lines[i],
	oldRp = (i ? rp : rp + ldp->patLines) - 1;
	
	sR = *rp;
	if( sR.right < sR.left ) {
		sR.left = sR.right;
		sR.right = rp->left;
	}
	if( sR.bottom < sR.top ) {
		sR.top = sR.bottom;
		sR.bottom = rp->top;
	}
	
	switch( ldp->drawMode )
	{
	case Ctrl_Line:
		MoveToEx( hdc, rp->left, rp->top, NULL );
		LineTo( hdc, rp->right, rp->bottom );
		break;
	
	case Ctrl_Rect:
		Rectangle( hdc, sR.left, sR.top, sR.right, sR.bottom );
		break;
	
	case Ctrl_Oval:
		Ellipse( hdc, sR.left, sR.top, sR.right, sR.bottom );
		break;
	
	case Ctrl_Lissajous:
		MoveToEx( hdc, oldRp->left, oldRp->top, NULL );
		LineTo(   hdc, rp->left, rp->top );
		MoveToEx( hdc, oldRp->right, oldRp->bottom, NULL );
		LineTo(   hdc, rp->right, rp->bottom );
		
		break;
	
	case Ctrl_Plot:
	{
		int			x = 4 * ldp->drawLine;
		int oldROP = SetROP2( hdc, R2_COPYPEN );
		HPEN oldPen;

		for( int j = 0;  j <= 1;  ++j ) {		/* draw twice, one repetition apart */
			if( i == 0 ) {
				HPEN oldPen = SelectPen( hdc, CreatePen(
					(oldROP == R2_COPYPEN) ? PS_SOLID : PS_DOT, 1, RGB(0,0,0)) );
				MoveToEx( hdc, x, -20000, NULL );
				LineTo(   hdc, x,  20000 );
				DeleteObject( SelectObject(hdc, oldPen) );
			}
			
			oldPen = SelectPen( hdc, CreatePen(PS_SOLID, 1, RGB(0xFF,0,0)) );
			MoveToEx( hdc, x-4, oldRp->top, NULL );
			LineTo(   hdc, x, rp->top );
			DeleteObject( SelectObject(hdc, oldPen) );
			
			oldPen = SelectPen( hdc, CreatePen(PS_DOT,   1, RGB(0xC0,0xC0,0)) );
			MoveToEx( hdc, x-4, oldRp->left, NULL );
			LineTo(   hdc, x, rp->left );
			DeleteObject( SelectObject(hdc, oldPen) );
			
			oldPen = SelectPen( hdc, CreatePen(PS_SOLID, 1, RGB(0,0xFF,0)) );
			MoveToEx( hdc, x-4, oldRp->bottom, NULL );
			LineTo(   hdc, x, rp->bottom );
			DeleteObject( SelectObject(hdc, oldPen) );
			
			oldPen = SelectPen( hdc, CreatePen(PS_DOT,   1, RGB(0,0xC0,0xC0)) );
			MoveToEx( hdc, x-4, oldRp->right, NULL );
			LineTo(   hdc, x, rp->right );
			DeleteObject( SelectObject(hdc, oldPen) );
			
			x += 4 * ldp->patLines;
		}
		
		SetROP2( hdc, oldROP );
		break;
	}
	}
}


/*	Draw the next line in a window.  Erase a line if enough have been drawn;
	all lines drawn will be erased.  If the current pattern is ended, make a
	new pattern and get ready to draw it next time.  EraseLine trails drawLine
	if erasing is enabled; otherwise it tracks drawLine.
*/
	static void
DrawWindow( HWND hwnd, HDC hdc, LinesData * ldp )
{
	if( ldp->drawLine < ldp->stopLine ) {
		//int oldROP = SetROP2( hdc, R2_XORPEN );
		short line = ldp->drawLine % ldp->patLines;
		HPEN oldPen = SelectPen( hdc, CreatePen(PS_SOLID, 1, ldp->colors[line]) );
		HBRUSH oldBrush = SelectBrush( hdc, GetStockObject(NULL_BRUSH) );
		
		DrawIt( hdc, ldp, line );

		DeleteObject( SelectObject(hdc, oldBrush) );
		DeleteObject( SelectObject(hdc, oldPen) );
		//SetROP2( hdc, oldROP );
	}
	
	if(   ldp->eraseLine >= 0  &&  ldp->eraseLine < ldp->stopLine
	   && ldp->eraseBehind )
	{
		//int oldROP = SetROP2( hdc, R2_XORPEN );
		short line = ldp->eraseLine % ldp->patLines;
		HPEN oldPen = SelectPen( hdc, CreatePen(PS_SOLID, 1, RGB(255,255,255)) );
		HBRUSH oldBrush = SelectBrush( hdc, GetStockObject(NULL_BRUSH) );
		
		DrawIt( hdc, ldp, line );

		DeleteObject( SelectObject(hdc, oldBrush) );
		DeleteObject( SelectObject(hdc, oldPen) );
		//SetROP2( hdc, oldROP );
	}
	
	++ldp->drawLine;
	++ldp->eraseLine;
	
	if( ldp->repeating ) {
		ldp->drawLine  %= ldp->stopLine;
		ldp->eraseLine %= ldp->stopLine;
	}
	
	if( (ldp->eraseBehind ? ldp->eraseLine : ldp->drawLine) > ldp->stopLine ) {
		RECT r;
		GetClientRect( hwnd, &r );
		HBRUSH whiteBrush = GetStockBrush( WHITE_BRUSH );
		FillRect( hdc, &r, whiteBrush );
		DeleteObject( whiteBrush );
		
		MakePattern( hwnd );
	}
}


/*	Fill the sine table.  Rather than use the Transcendental package, which
	would be slow, we use trig identities and a couple of pre-calculated
	values.  Got some use from TK!Solver.
*/
	static void
FillSine( void )
{
	for( int i = 0;  i <= SMAX/4;  ++i ) {
		double ipi = (double) i / SMAX * PI * 2;
		sine[i]        = sine[SMAX/2-i] =  sin(ipi);	/* symmetry of sine function */
		sine[SMAX/2+i] = sine[SMAX-i]   = -sine[i];
	}
}


	static BOOL
Cls_OnCreate( HWND hwnd, LPCREATESTRUCT csp )
{
	if( !sine[1] )
		FillSine();							/* sine table to trace lissajous figures */
	
	MDICREATESTRUCT *mcsp = (MDICREATESTRUCT *) csp->lpCreateParams;
	HWND cw = (HWND) mcsp->lParam;

	LinesData *ldp = (LinesData *) calloc( sizeof(LinesData), 1 );
	SetProp( hwnd, LINESDATAPROP, (HANDLE) ldp );
	ldp->eraseBehind = TRUE;
	ldp->repeating   = FALSE;
	ldp->drawMode    = Ctrl_Line;

	if( IsWindow(cw) ) {
		LinesData *cwldp = (LinesData *) GetProp( cw, LINESDATAPROP );
		
		if( cwldp ) {
			ldp->eraseBehind = cwldp->eraseBehind;
			ldp->repeating   = cwldp->repeating;
			ldp->drawMode    = cwldp->drawMode;
		}
	}

	TCHAR ttl[64];
	LoadString( hInst, ldp->drawMode - Ctrl_Line + IDS_LINESTTL1, ttl, chDIMOF(ttl) );
	SetWindowText( hwnd, ttl );

	SetTimer( hwnd, TIMERID, 16, NULL );

	MakePattern( hwnd );

	return true;
}


	static void
Cls_OnInitMenuPopup( HWND hwnd, HMENU hmenu, UINT item, BOOL fSystemMenu )
{
	if( hmenu == gFileMenu ) {
		EnableMenuItem( hmenu, ID_FILE_CLOSE, MF_BYCOMMAND | MF_ENABLED );
	}

	if( hmenu == gLinesMenu ) {
		LinesData *ldp = (LinesData *) GetProp( hwnd, LINESDATAPROP );
		
		EnableMenuItem( hmenu, ID_LINES_ERASEBEHIND, MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_LINES_REPEAT,      MF_BYCOMMAND | MF_ENABLED );
		if( ldp->repeating )
			EnableMenuItem( hmenu, ID_LINES_STEP,    MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_LINES_LINES,       MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_LINES_BOXES,       MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_LINES_OVALS,       MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_LINES_LISSAJOUS,   MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_LINES_PLOT,        MF_BYCOMMAND | MF_ENABLED );

		CheckMenuItem( hmenu, ID_LINES_ERASEBEHIND, MF_BYCOMMAND
			| (ldp->eraseBehind ? MF_CHECKED : MF_UNCHECKED) );
		CheckMenuItem( hmenu, ID_LINES_REPEAT, MF_BYCOMMAND
			| (ldp->repeating ? MF_CHECKED : MF_UNCHECKED) );
		CheckMenuRadioItem( hmenu, ID_LINES_LINES, ID_LINES_PLOT,
			ldp->drawMode - Ctrl_Line + ID_LINES_LINES, MF_BYCOMMAND );
	}
}


	static void
Cls_OnCommand( HWND hwnd, int id, HWND hwndCtl, UINT codeNotify )
{
	LinesData *ldp = (LinesData *) GetProp( hwnd, LINESDATAPROP );

	switch( id )
	{
	case ID_LINES_ERASEBEHIND:
		ldp->eraseBehind = ! ldp->eraseBehind;
		break;

	case ID_LINES_REPEAT:
		ldp->repeating = ! ldp->repeating;
		break;
	
	case ID_LINES_STEP:
	{
		HDC hdc = GetDC( hwnd );
		RECT r;
		GetClientRect( hwnd, &r );
		FillRect( hdc, &r, GetStockBrush(WHITE_BRUSH) );
		ReleaseDC( hwnd, hdc );
		
		MakePattern( hwnd );
		break;
	}

	case ID_LINES_LINES:
	case ID_LINES_BOXES:
	case ID_LINES_OVALS:
	case ID_LINES_LISSAJOUS:
	case ID_LINES_PLOT:
	{
		ldp->drawMode = id - ID_LINES_LINES + Ctrl_Line;
		
		TCHAR ttl[256];
		LoadString( hInst, ldp->drawMode - Ctrl_Line + IDS_LINESTTL1, ttl, chDIMOF(ttl) );
		SetWindowText( hwnd, ttl );
		ldp->eraseLine -= ldp->drawLine;
		ldp->drawLine = 0;
		break;
	}
	}
}


	static void
Cls_OnPaint( HWND hwnd )
{
	LinesData *ldp = (LinesData *) GetProp( hwnd, LINESDATAPROP );

	PAINTSTRUCT ps;
	HDC hdc = BeginPaint( hwnd, &ps );
	DrawWindow( hwnd, hdc, ldp );
	EndPaint( hwnd, &ps );
}


	static void
Cls_OnTimer( HWND hwnd, UINT id )
{
	if( id == TIMERID ) {
		LinesData *ldp = (LinesData *) GetProp( hwnd, LINESDATAPROP );
		if( ++ldp->delayCnt >= ldp->delay ) {
			ldp->delayCnt = 0;
			
			HDC hdc = GetDC( hwnd );
			DrawWindow( hwnd, hdc, ldp );
			ReleaseDC( hwnd, hdc );
		}
	}
}


	static void
Cls_OnDestroy( HWND hwnd )
{
	SendMessage( mainW, wmuChildClosing, (WPARAM) hwnd, 0 );
	LinesData *ldp = (LinesData *) RemoveProp( hwnd, LINESDATAPROP );
	if( ldp )
		free( (void *) ldp );
}


	static BOOL
DoCmdLineArg( HWND hwnd, TCHAR * arg )
{
	int len = strlen( arg );

	if( _tcsncicmp(arg, _T("/erase"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_ERASEBEHIND, 0 );
	else if( _tcsncicmp(arg, _T("/repeat"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_REPEAT, 0 );
	else if( _tcsncicmp(arg, _T("/step"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_STEP, 0 );
	else if( _tcsncicmp(arg, _T("/lines"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_LINES, 0 );
	else if( _tcsncicmp(arg, _T("/boxes"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_BOXES, 0 );
	else if( _tcsncicmp(arg, _T("/ovals"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_OVALS, 0 );
	else if( _tcsncicmp(arg, _T("/lissajous"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_LISSAJOUS, 0 );
	else if( _tcsncicmp(arg, _T("/plot"), len) == 0 )
		SendMessage( hwnd, WM_COMMAND, ID_LINES_PLOT, 0 );
	else
		return FALSE;

	return TRUE;
}


	LRESULT CALLBACK
LinesWndProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	HANDLE_MSG( hwnd, WM_CREATE,        Cls_OnCreate );
	HANDLE_MSG( hwnd, WM_COMMAND,       Cls_OnCommand );
	HANDLE_MSG( hwnd, WM_INITMENUPOPUP, Cls_OnInitMenuPopup );
	HANDLE_MSG( hwnd, WM_PAINT,	        Cls_OnPaint );
	HANDLE_MSG( hwnd, WM_TIMER,         Cls_OnTimer );
	HANDLE_MSG( hwnd, WM_DESTROY,       Cls_OnDestroy );

	case WM_MDIACTIVATE:
		PostMessage( mainW, wmuChildActivate, 0, 0 );
		break;

	case wmuCmdLineArg:
		return DoCmdLineArg( hwnd, (TCHAR *) wParam );
	}

	return DefMDIChildProc( hwnd, uMsg, wParam, lParam );
}
